﻿// -------------------------------------------------------------------------------------
//
// Explorer.cs                                                   Ulrich Müller  26.05.08
//                                                               PCS1700 C# 2005 Express
// Klasse zum Betrieb des ROBO Explorers
// FishFace2005.DLL wird genutzt, Anwendung using FishFace40;
//
// -------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using FishFace40;

namespace Explorer5 {

  public struct DatenEingaenge {
    public bool SpurL;
    public bool SpurR;
    public int Distance;
    public int Temperatur;
    public int Farbe;
    public int Helligkeit;
  }
  public struct Farben { 
    public int blauMin;
    public int blauMax;
    public int gruenMin;
    public int gruenMax;
    public int rotMin;
    public int rotMax;
  }

  class Explorer {

    // --- private globale Variable ---------------

    FishFace ft = new FishFace();
    Thread   MessungEingaenge;
    Thread   TemperaturMelder;
    Thread   LichtSchaltung;
    Thread   FarbErkennung;
    Thread   SpurFahren;

    Speed geschwindigkeit  = Speed.Full;
    DatenEingaenge messung = new DatenEingaenge();
    int heissMax           = 400;
    int dunkelMin          = 500;
    int disMin             = 12;
    Farben grenzFarben     = new Farben();
    bool aufSpur           = true;
    bool tempStop          = false;
    bool lichtStop         = false;
    bool farbStop          = false;
    bool spurStop          = false;

    // --- Belegung des ROBO Interfaces an USB ------------

    const Out mLinks   = Out.M2;
    const Out mRechts  = Out.M1;
    const Dir cVor     = Dir.Left;
    const Dir cRuck    = Dir.Right;
    const Dir cOff     = Dir.Off;

    const Out oLampeR  = Out.O5;
    const Out oLampeL  = Out.O6;
    const Out oLampeT  = Out.O7;
    const Out oHupe    = Out.O8;

    const Inp iImpulsL = Inp.I1;
    const Inp iSpurR   = Inp.I3;
    const Inp iSpurL   = Inp.I4;

    const Inp dUltra   = Inp.D1;

    const Inp aLicht   = Inp.AX;
    const Inp aTemp    = Inp.AY;

    const Inp aFarbe   = Inp.A1;

    public delegate void Messwerte(object sender, DatenEingaenge Werte);
    public event         Messwerte WerteEingaenge;

    public Explorer() {
      grenzFarben.blauMin  = 240; grenzFarben.blauMax  = 260;
      grenzFarben.gruenMin = 200; grenzFarben.gruenMax = 220;
      grenzFarben.rotMin   = 170; grenzFarben.rotMax   = 190;
    }

    public void OpenExplorer(IFTypen iTyp, int SerialNr) {
      ft.OpenInterface(iTyp, SerialNr,Distance.UltraSonic);
      if (WerteEingaenge != null) {
        MessungEingaenge = new Thread(MessLoop);
        MessungEingaenge.IsBackground = true;
        MessungEingaenge.Start();
      }
    }
    public void OpenExplorer() {
      OpenExplorer(IFTypen.ftROBO_first_USB, 0);
    }
    public void CloseExplorer() {
      ft.NotHalt = true;
      if(WerteEingaenge != null) MessungEingaenge.Join();
      StopTemperaturUeberwachung();
      StopLichtUeberwachung();
      StopFarbUeberwachung();
      ft.CloseInterface();
    }

    // --- Eigenschaften -----

    public Speed          Geschwindigkeit {
      get { return geschwindigkeit; }
      set { geschwindigkeit = value; }
    }
    public int            MinimalDistance {
      get { return disMin; }
      set { disMin = value; }
    }
    public DatenEingaenge MessDaten { 
      get {
        messung.Distance   = ft.GetDistance(dUltra);
        messung.Farbe      = ft.GetVoltage(aFarbe);
        messung.Helligkeit = ft.GetAnalog(aLicht);
        messung.SpurL      = !ft.GetInput(iSpurL);
        messung.SpurR      = !ft.GetInput(iSpurR);
        messung.Temperatur = ft.GetAnalog(aTemp);
        return messung;        
      }
    }
    public int            HeissMax {
      get { return heissMax; }
      set { heissMax = value; }
    }
    public int            DunkelMin { 
      get { return dunkelMin; } 
      set { dunkelMin = value; } }
    public Farben         FarbGrenzen { 
      get { return grenzFarben; } 
      set { grenzFarben = value; } 
    }

    // --- Methoden : Fahren ---

    public void Stopp() {
      ft.SetMotor(mLinks, cOff);
      ft.SetMotor(mRechts, cOff);
    }
    public void StoppRechts(Speed Geschwindigkeit) {
      ft.SetMotor(mLinks, cVor, Geschwindigkeit);
      ft.SetMotor(mRechts, cOff);
    }
    public void StoppRechts() {
      StoppRechts(geschwindigkeit);
    }
    public void StoppLinks(Speed Geschwindigkeit) {
      ft.SetMotor(mLinks, cOff);
      ft.SetMotor(mRechts, cVor, Geschwindigkeit);
    }
    public void StoppLinks() {
      StoppLinks(geschwindigkeit);
    }
    public void Vorwaerts(Speed Geschwindigkeit) {
      ft.SetMotor(mLinks, cVor, Geschwindigkeit);
      ft.SetMotor(mRechts, cVor, Geschwindigkeit);
    }
    public void Vorwaerts() {
      Vorwaerts(geschwindigkeit);
    }
    public void Rueckwaerts(Speed Geschwindigkeit) {
      ft.SetMotor(mLinks, cRuck, Geschwindigkeit);
      ft.SetMotor(mRechts, cRuck, Geschwindigkeit);
    }
    public void Rueckwaerts() {
      Rueckwaerts(geschwindigkeit);
    }
    public void DrehenLinks(Speed Geschwindigkeit) {
      ft.SetMotor(mLinks, cRuck, Geschwindigkeit);
      ft.SetMotor(mRechts, cVor, Geschwindigkeit);
    }
    public void DrehenLinks() {
      DrehenLinks(geschwindigkeit);
    }
    public void DrehenRechts(Speed Geschwindigkeit) {
      ft.SetMotor(mLinks, cVor, Geschwindigkeit);
      ft.SetMotor(mRechts, cRuck, Geschwindigkeit);
    }
    public void DrehenRechts() {
      DrehenRechts(geschwindigkeit);
    }

    // --- Methoden : Erweitertes Fahren -----

    public bool HindernisErkennung() { 
      if(ft.GetDistance(dUltra) < disMin) return false;
      Stopp();
      ft.Pause(1000);
      Rueckwaerts(Speed.Half);
      ft.WaitForGreater(dUltra, disMin);
      Stopp();
      DrehenLinks(Speed.Half);
      ft.Pause(500);
      return true;
    }
    public void Spursuche() {
      int Zahler;
      aufSpur = false;
      while (!ft.Finish()) {
        Zahler = 52;
        DrehenLinks();
        do {
          if (!ft.GetInput(iSpurL) || !ft.GetInput(iSpurR)) {
            Stopp(); aufSpur = true; return;
          }
          ft.WaitForHigh(iImpulsL);
          if (HindernisErkennung()) { Zahler = 52; DrehenLinks(); }
          else Zahler--;
        } while (Zahler > 0 && !ft.Finish());
        Zahler = 10;
        Vorwaerts(Speed.Half);
        do {
          if (!ft.GetInput(iSpurL) || !ft.GetInput(iSpurR)) {
            Stopp(); aufSpur = true; return;
          }
          ft.WaitForHigh(iImpulsL);
          if (HindernisErkennung()) Zahler = 0;
          else Zahler--;
        } while (Zahler > 0 && !ft.Finish());
      }
    }

    // --- Methoden : Licht, Hupe ---

    public void  LichtAn() {
      ft.SetLamp(oLampeL, Dir.On);
      ft.SetLamp(oLampeR, Dir.On);
    }
    public void  LichtAus() {
      ft.SetLamp(oLampeL, Dir.Off);
      ft.SetLamp(oLampeR, Dir.Off);
    }
    public void  HupeEin() { ft.SetLamp(oHupe, Dir.On); }
    public void  HupeAus() { ft.SetLamp(oHupe, Dir.Off); }
    private void Hupen(int nMal, int Dauer) {
      for (int i = 0; i < nMal; i++) {
        HupeEin();
        ft.Pause(Dauer);
        HupeAus();
        ft.Pause(Dauer);
      }
    }

    // --- Thread Routinen ---

    private void MessLoop() {
      while (!ft.Finish()) {
        WerteEingaenge(this, MessDaten);
        ft.Pause(111);
      }
    }
    public void  StartTemperaturUeberwachung() {
      TemperaturMelder = new Thread(TempLoop);
      TemperaturMelder.IsBackground = true;
      tempStop = false;
      TemperaturMelder.Start();
    }
    public void  StopTemperaturUeberwachung() {
      if (TemperaturMelder != null) {
        tempStop = true;
        TemperaturMelder.Join();
      }
    }
    private void TempLoop() {
      while (!tempStop && !ft.Finish()) {
        if ((ft.GetAnalog(aTemp)) < heissMax) {
          ft.SetLamp(oLampeT, Dir.On);
          ft.Pause(100);
          ft.SetLamp(oLampeT, Dir.Off);
          ft.Pause(600);
        }
      }
    }
    public void  StartLichtUeberwachung() {
      LichtSchaltung = new Thread(LichtLoop);
      LichtSchaltung.IsBackground = true;
      lichtStop = false;
      LichtSchaltung.Start();
    }
    public void  StopLichtUeberwachung() {
      if (LichtSchaltung != null) {
        lichtStop = true;
        LichtSchaltung.Join();
      }
    }
    private void LichtLoop() {
      while (!lichtStop && !ft.Finish()) {
        if (ft.GetAnalog(aLicht) > dunkelMin) LichtAn();
        else LichtAus();
      }
    }
    public void  StartFarbUeberwachung() {
      FarbErkennung = new Thread(FarbLoop);
      FarbErkennung.IsBackground = true;
      farbStop = false;
      FarbErkennung.Start();
    }
    public void  StopFarbUeberwachung() {
      if (FarbErkennung != null) {
        farbStop = true;
        FarbErkennung.Join();
      }
    }
    private void FarbLoop() {
      int farbe;
      while (!farbStop && !ft.Finish()) {
        if (aufSpur) {
          farbe = ft.GetVoltage(aFarbe);
          if (farbe > grenzFarben.blauMin && farbe < grenzFarben.blauMax) 
            Hupen(4, 100);
          if (farbe > grenzFarben.gruenMin && farbe < grenzFarben.gruenMax) 
            Hupen(3, 300);
          if (farbe > grenzFarben.rotMin && farbe < grenzFarben.rotMax)
            Hupen(2, 600);
        }
      }
    }
    public void  StartSpurfahren() {
      SpurFahren = new Thread(SpurLoop);
      SpurFahren.IsBackground = true;
      spurStop = false;
      SpurFahren.Start();
    }
    public void  StopSpurfahren() {
      if (SpurFahren != null) {
        spurStop = true;
        SpurFahren.Join();
      }
    }
    private void SpurLoop() {
      while (!spurStop && !ft.Finish()) {
        switch (ft.GetInputs() & 0xC) { 
          case 0xC:
            Stopp(); Spursuche(); break;
          case 0x8:
            StoppLinks(); HindernisErkennung(); break;
          case 0x4:
            StoppRechts(); HindernisErkennung(); break;
          default: Vorwaerts(Speed.Half); HindernisErkennung(); break;
        }
      }
    }
  }
}
